# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from libcanonical.environ import defaults


LOGGING_CONFIG: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "libcanonical.utils.logging.DefaultFormatter",
            "fmt": "%(asctime)s %(levelprefix)s %(message)s",
            "use_colors": True,
            'datefmt': '%Y-%m-%d %H:%M:%S.%f'
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {
        "canonical": {"handlers": ["default"], "level": defaults.loglevel, "propagate": True},
        "canonical.error": {"level": "INFO"},
    },
}


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Optional
from typing import TypedDict


LoggerConfigDict = TypedDict('LoggerConfigDict', {
    'handlers': list[str],
    'level': str,
    'propagate': bool
})


class LoggingConfigDict(TypedDict):
    version: int
    disable_existing_loggers: bool
    filters: Optional[dict[str, Any]]
    loggers: dict[str, LoggerConfigDict]


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import binascii
import json
from typing import Any
from typing import TypeVar

import pydantic


T = TypeVar('T')


def bytes_to_number(value: bytes) -> int: # pragma: no cover
    return int(binascii.b2a_hex(value), 16)


def b64decode(buf: bytes | str, decoder: type[T] = bytes) -> T:
    """Decode the given byte-sequence or string."""
    adapter = pydantic.TypeAdapter(decoder)
    if isinstance(buf, str):
        buf = buf.encode("ascii")
    rem = len(buf) % 4
    if rem > 0:
        buf += b"=" * (4 - rem)
    return adapter.validate_python(base64.urlsafe_b64decode(buf))


def b64decode_int(value: bytes | str) -> int:
    return bytes_to_number(b64decode(value))


def b64decode_json(
    buf: bytes | str,
    encoding: str = 'utf-8',
    require: type[list[Any]] | type[dict[str, Any]] | None = None
) -> dict[str, Any] | list[Any]:
    """Deserialize a Base64-encoded string or byte-sequence as JSON."""
    result = json.loads(bytes.decode(b64decode(buf), encoding))
    if not isinstance(result, (require,) if require else (dict, list)): # pragma: no cover
        raise ValueError 
    return result


def b64encode_int(
    value: int,
    encoder: type[T] = bytes
) -> T: # pragma: no cover
    return b64encode(int.to_bytes(value, (value.bit_length() + 7) // 8, 'big'), encoder=encoder)


def b64encode_json(
    obj: dict[str, Any] | list[Any],
    encoder: type[T] = bytes
) -> T:
    """Encode the given dictionary as JSON and return the Base64-encoded
    byte-sequence.
    """
    return b64encode(json.dumps(obj, sort_keys=True), 'utf-8', encoder=encoder)


def b64encode(
    buf: bytes | str,
    encoding: str = 'utf-8',
    encoder: type[T] = bytes
) -> T:
    """Encode the given string or byte-sequence using the specified
    encoding.
    """
    adapter = pydantic.TypeAdapter(encoder)
    if isinstance(buf, str):
        buf = str.encode(buf, encoding=encoding)
    value = base64.urlsafe_b64encode(buf).replace(b"=", b"")
    return adapter.validate_python(value)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical import StringType


class ConstrainedStringType(StringType):
    pattern = r'^[0-9]+$'
    min_length = 2
    max_length = 3


@pytest.mark.parametrize('v', [
    'aaa',
    '1',
    '1234'
])
def test_invalid_model(v: str):
    class Model(pydantic.BaseModel):
        value: ConstrainedStringType

    with pytest.raises(pydantic.ValidationError):
        Model.model_validate({'value': v})


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class Request(pydantic.BaseModel):
    host: str = pydantic.Field(
        default=...,
        description=(
            "The IP address of the client, as seen by the "
            "server."
        )
    )

    url: str = pydantic.Field(
        default=...,
        description="The full URL that was requested."
    )


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pkg_resources
import warnings
from typing import Any


def get_entrypoints(name: str):
    entrypoints = [
        (entry_point.name, entry_point.load())
        for entry_point
        in pkg_resources.iter_entry_points(name)
    ]
    return sorted(entrypoints, key=lambda x: getattr(x[1], 'WEIGHT', 0))


def run_entrypoints(
    name: str,
    fn: str | None = None,
    namespace: str = 'canonical',
    *args: Any,
    **kwargs: Any
) -> None:
    """Runs the entrypoints specified by `name`."""
    for _, entry_point in get_entrypoints(name):
        f = entry_point if not fn else getattr(entry_point, fn, None)
        if not callable(f):
            warnings.warn(
                f"Entrypoint {entry_point} (fn: {fn}) was not callable."
            )
            continue
        f(*args, **kwargs)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import pathlib

import pydantic


class EnvironmentVariables(pydantic.BaseModel):
    name: str = pydantic.Field(
        default=...
    )

    loglevel: str = pydantic.Field(
        alias='LOGLEVEL',
        default='INFO'
    )

    vardir: pathlib.Path = pydantic.Field(
        alias='APP_VARDIR',
        default=...
    )

    @pydantic.model_validator(mode='before')
    def validate_params(cls, params: dict[str, str]):
        name = params.get('name')
        if not isinstance(name, str):
            raise pydantic.ValidationError()
        if not params.get('APP_VARDIR'):
            params['APP_VARDIR'] = f'/var/lib/{name}'
        return params

    @classmethod
    def parse(cls, app_name: str):
        return cls.model_validate({**os.environ, 'name': app_name})


defaults = EnvironmentVariables.parse('.canonical')


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler

from libcanonical.utils.encoding import b64decode
from libcanonical.utils.encoding import b64encode


__all__: list[str] = [
    'Base64'
]

T = TypeVar('T', bound='Base64')


class Base64(bytes):
    __module__: str = 'libcanonical.types'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.union_schema([
                    core_schema.is_instance_schema(bytes),
                    core_schema.chain_schema([
                        core_schema.is_instance_schema(str),
                        core_schema.no_info_plain_validator_function(cls.b64decode),
                        core_schema.no_info_plain_validator_function(cls.validate)
                    ])
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.b64encode
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def b64decode(cls, value: str):
        return b64decode(value)

    @classmethod
    def b64encode(cls, value: bytes | str) -> str:
        return  b64encode(value, encoder=str)

    @classmethod
    def validate(cls, instance: T) -> T:
        return instance


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


T = TypeVar('T')


class StringType(str):
    """Base class for string types."""
    __module__: str = 'libcanonical.types'
    lowercase: bool = False
    max_length: int | None = None
    min_length: int | None = None
    openapi_title: str | None = None
    openapi_format: str | None = None
    patterns: re.Pattern[Any] | list[re.Pattern[Any]] = []
    pattern: str | None = None
    strip_whitespace: bool = True

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=cls.__default_schema__(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    cls.__default_schema__(),
                    core_schema.no_info_plain_validator_function(cls.transform),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(cls.__default_schema__())

    @classmethod
    def __default_schema__(cls):
        return core_schema.str_schema(
            max_length=cls.max_length,
            min_length=cls.min_length,
            pattern=cls.pattern,
            strip_whitespace=cls.strip_whitespace,
            to_lower=cls.lowercase
        )

    @classmethod
    def transform(cls, v: str) -> str:
        return v

    @classmethod
    def validate(cls, v: str) -> str:
        return cls(v)

    @classmethod
    def validate_pattern(cls, v: Any, _: Any = None) -> str:
        if not isinstance(v, str):
            raise ValueError(f"{cls.__name__} must be instantiated from a string type.")
        patterns = cls.patterns
        if not isinstance(patterns, list): # pragma: no cover
            patterns = [patterns]
        for pattern in patterns:
            if not pattern.match(v):
                raise ValueError(f"not a valid {cls.__name__}.")
        return v

    def __new__(cls, object: object):
        self = super().__new__(cls, object=object)
        self.type_post_init()
        return self

    def type_post_init(self):
        pass

    def __repr__(self) -> str: # pragma: no cover
        return f'<{type(self).__name__}: {str(self)}>'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Sequence
from typing import Generator
from typing import TypeVar

T = TypeVar('T')


def chunks(iterable: Sequence[T], n: int = 1) -> Generator[list[T], None, None]:
    l = len(iterable)
    for ndx in range(0, l, n):
        yield list(iterable[ndx:min(ndx + n, l)])


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

from .stringtype import StringType


class HTTPResourceLocator(StringType):
    __module__: str = 'libcanonical.types'
    max_length = 2048

    @classmethod
    def validate(cls, v: str, _: Any = None):
        p = urllib.parse.urlparse(v)
        if p.scheme not in {'http', 'https'}:
            raise ValueError(f"Not a valid URL: {v[:128]}")
        return cls(v)

    @property
    def query(self) -> dict[str, str | list[str]]:
        p =urllib.parse.urlparse(self)
        values: dict[str, list[str] | str] = {}
        for name, value in urllib.parse.parse_qs(p.query).items():
            if len(value) == 1:
                value = value[0]
            values[name] = value
        return values

    def with_query(self, **kwargs: Any) -> 'HTTPResourceLocator':
        p: list[str] = list(urllib.parse.urlparse(self))
        q = dict(urllib.parse.parse_qs(p[4]))
        q.update(kwargs)
        p[4] = urllib.parse.urlencode(q, doseq=True)
        return HTTPResourceLocator(urllib.parse.urlunparse(p))
        


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical import Phonenumber



class Model(pydantic.BaseModel):
    phonenumber: Phonenumber


@pytest.mark.parametrize('value,exc_class', [
    ('+31612345678', None),
    ('+31456', pydantic.ValidationError),
    ('a', pydantic.ValidationError),
    (1, pydantic.ValidationError)
])
def test_validate(value: str, exc_class: type[BaseException] | None):
    adapter = pydantic.TypeAdapter(Phonenumber)
    if exc_class is None:
        adapter.validate_python(value)
    else:
        with pytest.raises(exc_class):
            adapter.validate_python(value)


@pytest.mark.parametrize('value', [
    '+31612345678'
])
def test_validate_model(value: str):
    obj = Model.model_validate({'phonenumber': value})
    assert isinstance(obj.phonenumber, Phonenumber)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic



EncodingLiteral = Literal[
    'base64',
    'hex',
    'utf-8',
]


KindLiteral = Literal[
    'body',
    'headers',
    #'json',
    'query',
]


class HTTPRequestRef(pydantic.BaseModel):
    kind: KindLiteral = pydantic.Field(
        default=...,
        description=(
            "Specifies the component of the HTTP request that "
            "is being referenced."
        )
    )

    name: str = pydantic.Field(
        default='',
        description=(
            "Indicates where to find the the referent.\n\n"
            "- For `body`, this field is ignored.\n\n"
            "- For `headers` and `query`, the `name` field."
            "indicates the name of the header or query parameter.\n\n"
            "Note that when `kind=query`, repeated query parameters "
            "are parsed as a list and the `encoding` is applied to "
            "each member in the list."
        )
    )

    encoding: EncodingLiteral = pydantic.Field(
        default=...,
        description=(
            "The content encoding of the referent."
        )
    )

    def model_post_init(self, _: Any) -> None:
        if self.kind != 'body' and not self.name:
            raise ValueError("The `.name` field is required.")


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import contextlib
import inspect
import logging.config
import signal
import time
import os
from typing import cast
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Coroutine
from typing import TypeVar

from libcanonical.environ import EnvironmentVariables
from libcanonical.utils import logger
from libcanonical.utils.logging import LoggingConfigDict
from ._entrypoint import run_entrypoints
from ._logging import LOGGING_CONFIG


R = TypeVar('R')


class MainProcess:
    """The main process of an application."""
    environment: type[EnvironmentVariables] = EnvironmentVariables
    handle_signals: bool = True
    interval: float = 0.01
    logger = logger
    log_tracebacks: bool = True
    max_age: int = 0
    min_runtime: int = 0
    must_drop: bool = True
    must_exit: bool = False
    must_reload: bool = False
    started: int
    suspended: bool = False
    uid: int
    _step: int = 0

    class ConfigurationFailure(Exception):

        def __init__(self, message: str):
            self.message = message

    @property
    def step(self) -> int:
        return self._step

    @classmethod
    def run(cls, name: str, **kwargs: Any):
        self = cls(name=name, **kwargs)
        self.main()

    @property
    def age(self) -> int:
        return int(time.monotonic() - self.started)

    @property
    def runtime(self) -> int:
        return int(time.monotonic() - self.started)

    def __init__(
        self, *,
        name: str,
        **kwargs: Any
    ):
        self.name = name
        self.environ = self.environment.parse(self.name)
        self.timestamp = time.monotonic()

    def configure(self, reloading: bool = False) -> None | Coroutine[None, None, None]:
        """A hook to configure and setup the process, prior to entering the
        main event loop. This method is also invoked when the application is
        requested to reload using ``SIGHUP``.

        Args:
            reloading (bool): indicates if this method was invoked during
                a reload. If `reloading` is ``False``, then the main event
                loop is not running.

        Returns:
            None
        """
        pass

    def get_logging_config(self) -> LoggingConfigDict:
        return cast(LoggingConfigDict, LOGGING_CONFIG)

    def initialize(self) -> None:
        """Initializes the application on the current system."""
        return

    def is_initialized(self):
        """Return a boolean indicating if the application is initialized."""
        return all([
            os.path.exists(self.environ.vardir)
        ])

    def is_runlimit_reached(self) -> bool:
        """Hook to indicate that the run limit is reached and the process
        must exit.
        """
        return False

    def is_stopped(self):
        return self.must_exit

    def is_suspended(self):
        return self.suspended

    def log_step(self, duration: float):
        self.logger.debug(
            "Completed main event in %.02fs (step: %s)",
            duration,
            self.step
        )

    def main(self) -> None:
        if self.handle_signals:
            self._bind_signal_handlers()
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self.__main__())

    def main_event(self) -> None | Coroutine[None, None, None]:
        """The main event of the process. Subclasses must override
        this method.
        """
        raise NotImplementedError

    def on_completed(self):
        """A hook that is invoked when the process exits succesfully."""
        return

    def on_sigint(self) -> None:
        """Invoked when the process receives a ``SIGINT`` signal."""
        self.logger.info("Caught SIGINT (pid: %s)", os.getpid())
        if self.must_exit:
            # If we already received a SIGINT, assume that the
            # process is not existing and the user wants to
            # kill it off.
            os.kill(os.getpid(), signal.SIGKILL)
        self.must_exit = True

    def on_sighup(self) -> None:
        """Invoked when the process receives a ``SIGHUP`` signal."""
        self.must_reload = True

    def on_sigterm(self) -> None:
        """Invoked when the process receives a ``SIGTERM`` signal."""
        # TODO: Implement this as must_kill
        self.must_exit = True

    def on_sigusr1(self) -> None:
        """Hook to handle ``SIGUSR1``. Subclasses may override this method
        without calling :func:`super()`.
        """
        pass

    def on_sigusr2(self) -> None:
        """Hook to handle ``SIGUSR2``. Subclasses may override this method
        without calling :func:`super()`.
        """
        pass

    def setup_logging(self):
        """Configures the loggers for the application."""
        logging.config.dictConfig(dict(self.get_logging_config()))

    def stop(self):
        self.must_exit = True

    @contextlib.asynccontextmanager
    async def suspend(self, duration: float):
        if self.is_suspended():
            raise RuntimeError("Cannot suspend while suspended.")
        try:
            self.suspended = True
            yield
            elapsed = 0.0
            while elapsed < duration:
                await asyncio.sleep(0.1)
                if self.must_exit or self.must_reload:
                    self.suspended = False
                    break
                elapsed += 0.1
        finally:
            self.suspended = False

    def teardown(self) -> None | Coroutine[None, None, None]:
        """Called during application teardown. Subclasses may override
        this method without calling :func:`super()`.
        """
        pass

    async def __main__(self) -> None:
        self.started = int(time.monotonic())
        if not (await self._run(self._configure)):
            self.logger.critical("Failed configuration, exiting main event loop.")
            await asyncio.sleep(5)
            return
        self.logger.debug("Starting main event loop")
        if self.uid == 0:
            self.logger.warning("This process is running as a privileged user.")
        while True:
            t0 = time.time()
            if self.is_runlimit_reached():
                self.stop()
            if self.must_exit:
                self.logger.info(
                    "Shutting down %s (pid: %s)",
                    self.name,
                    self.pid
                )
                await self._teardown()
                break
            if self.must_reload and not await self._run(self._configure, reloading=True):
                self.stop()
                continue
            self._step += 1
            try:
                await self._run(self.main_event)
            except Exception as e:
                self._log_exception(e)
                await asyncio.sleep(1)
            finally:
                duration = time.time() - t0
                self.log_step(duration)
                t = self.min_runtime - duration
                if t > 0:
                    async with self.suspend(t):
                        self.logger.debug("Sleeping for %s seconds.", t)
                else:
                    await asyncio.sleep(max(self.interval - duration, 0.0))

    def _bind_signal_handlers(self):
        signal.signal(signal.SIGHUP, lambda s, f: self.on_sighup())
        signal.signal(signal.SIGINT, lambda s, f: self.on_sigint())
        signal.signal(signal.SIGUSR1, lambda s, f: self.on_sigusr1())
        signal.signal(signal.SIGUSR2, lambda s, f: self.on_sigusr2())

    def _initialize(self):
        if not os.access(self.environ.vardir.parent, os.W_OK):
            self.logger.critical(
                "Unable to create state directory %s",
                self.environ.vardir
            )
            self.stop()
        if self.is_stopped():
            return
        if not os.path.exists(self.environ.vardir):
            os.makedirs(self.environ.vardir, exist_ok=True)
        self.initialize()

    def _log_exception(self, e: BaseException, *args: Any, **kwargs: Any):
        self.logger.error(
            "Caught %s",
            type(e).__name__
        )
        if self.log_tracebacks:
            self.logger.exception("The traceback was:")

    async def _configure(self, reloading: bool = False):
        if not reloading:
            run_entrypoints('canonical')
        if reloading:
            self.logger.warning("Reloading %s (pid: %s)", self.name, self.pid)
        self.setup_logging()
        self.uid = os.getuid()
        self.pid = os.getpid()
        if not reloading and not self.is_initialized():
            self._initialize()
        try:
            await self._run(self.configure, reloading=reloading)
            success = True
        except self.ConfigurationFailure as e:
            self.logger.critical(e.message)
            success = False
        except Exception as e:
            self._log_exception(e)
            success = False
        finally:
            self.must_reload = False
        return success

    async def _run(
        self,
        func: Callable[..., R | Awaitable[R]],
        *args: Any,
        **kwargs: Any
    ):
        result = func(*args, **kwargs)
        if inspect.iscoroutinefunction(func):
            assert inspect.isawaitable(result)
            result = await result
        return result

    async def _teardown(self):
        try:
            await self._run(self.teardown)
            await self._run(self.on_completed)
        except Exception as e:
            self._log_exception(e)


#!/usr/bin/env python3
import json
import os
import pathlib
from setuptools import find_namespace_packages
from setuptools import setup


SETUPDIR = pathlib.Path(__file__).parent

NAMESPACE = 'canonical.ext'

packages = find_namespace_packages(
    where=SETUPDIR,
    exclude={'build', 'dist', 'tests', 'var'}
)
opts = json.loads((open(SETUPDIR.joinpath(packages[0], 'package.json')).read()))
version = str.strip(open(SETUPDIR.joinpath('VERSION')).read())
if os.path.exists(SETUPDIR.joinpath('README.md')):
    with open(SETUPDIR.joinpath('README.md'), encoding='utf-8') as f:
        opts['long_description'] = f.read()
        opts['long_description_content_type'] = "text/markdown"

setup(
    version=version,
    packages=packages,
    include_package_data=True,
    **opts)



# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging


class TracebackInfoFilter(logging.Filter):
    """Clear or restore the exception on log records"""

    def __init__(self, clear: bool = True):
        self.clear = clear

    def filter(self, record: logging.LogRecord):
        if self.clear:
            record.exc_info = None
            record.exc_text = None
        return True


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import collections.abc
import datetime
import json
import logging
import sys
from copy import copy
from typing import Callable
from typing import Literal

import click
import pydantic

TRACE_LOG_LEVEL = 5


def format_colors(color: str) -> Callable[[str], str]:
    return lambda level_name: click.style(level_name, fg=color)


class ColourizedFormatter(logging.Formatter):
    """A custom log formatter class that:

    * Outputs the LOG_LEVEL with an appropriate color.
    * If a log call includes an `extra={"color_message": ...}` it will be used
      for formatting the output, instead of the plain text message.
    """
    __module__: str = 'libcanonical.utils.logging'

    level_name_colors: dict[int, Callable[..., str]] = {
        TRACE_LOG_LEVEL: format_colors("blue"),
        logging.DEBUG: format_colors("cyan"),
        logging.INFO: format_colors("green"),
        logging.WARNING: format_colors("yellow"),
        logging.ERROR: format_colors("red"),
        logging.CRITICAL: format_colors("bright_red"),
    }

    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: Literal["%", "{", "$"] = "%",
        use_colors: bool | None = None,
    ):
        if use_colors in (True, False):
            self.use_colors = use_colors
        else:
            self.use_colors = sys.stdout.isatty()
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def color_level_name(self, level_name: str, level_no: int) -> str:
        def default(level_name: str) -> str:
            return str(level_name)  # pragma: no cover

        func = self.level_name_colors.get(level_no, default)
        return func(level_name)

    def should_use_colors(self) -> bool:
        return True  # pragma: no cover

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None):
        if not datefmt:
            return super().formatTime(record, datefmt=datefmt)

        return datetime.datetime.fromtimestamp(record.created).astimezone().strftime(datefmt)

    def formatMessage(self, record: logging.LogRecord) -> str:
        recordcopy = copy(record)
        levelname = recordcopy.levelname
        seperator = " " * (8 - len(recordcopy.levelname))
        message = recordcopy.getMessage()
        if isinstance(message, pydantic.BaseModel):
            message = message.model_dump(mode='json')
        if isinstance(message, collections.abc.Mapping):
            if message.get('message'):
                message = message['message']
            else:
                message = json.dumps(message, indent=2)

        if self.use_colors:
            levelname = self.color_level_name(levelname, recordcopy.levelno)
            if "color_message" in recordcopy.__dict__:
                recordcopy.msg = recordcopy.__dict__["color_message"]
                recordcopy.__dict__["message"] = message
        recordcopy.__dict__["levelprefix"] = levelname + ":" + seperator
        return super().formatMessage(recordcopy)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import http
import logging
from copy import copy
from typing import Callable

import click

from ._colourizedformatter import ColourizedFormatter
from ._colourizedformatter import format_colors


class AccessFormatter(ColourizedFormatter):
    __module__: str = 'libcanonical.utils.logging'
    status_code_colours: dict[int, Callable[..., str]] = {
        1: format_colors("bright_white"),
        2: format_colors("green"),
        3: format_colors("yellow"),
        4: format_colors("red"),
        5: format_colors("bright_red"),
    }

    def get_status_code(self, status_code: int) -> str:
        try:
            status_phrase = http.HTTPStatus(status_code).phrase
        except ValueError:
            status_phrase = ""
        status_and_phrase = f"{status_code} {status_phrase}"
        if self.use_colors:

            def default(code: int) -> str:
                return status_and_phrase  # pragma: no cover

            func = self.status_code_colours.get(status_code // 100, default)
            return func(status_and_phrase)
        return status_and_phrase

    def formatMessage(self, record: logging.LogRecord) -> str:
        client_addr: str
        method: str
        full_path: str
        http_version: str
        status_code: str
        recordcopy = copy(record)
        (
            client_addr,
            method,
            full_path,
            http_version,
            status_code,
        ) = recordcopy.args  # type: ignore[misc]
        status_code = self.get_status_code(int(status_code))  # type: ignore[arg-type]
        request_line = f"{method} {full_path} HTTP/{http_version}"
        if self.use_colors:
            request_line = click.style(request_line, bold=True)
        recordcopy.__dict__.update(
            {
                "client_addr": client_addr,
                "request_line": request_line,
                "status_code": status_code,
            }
        )
        return super().formatMessage(recordcopy)
